
# MIT License

# Copyright (c) 2023 Markus Iser, Karlsruhe Institute of Technology (KIT)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

import pandas as pd
import os
import glob
import warnings

from gbd_core.contexts import suffixes, identify, get_context_by_suffix
from gbd_core.api import GBD, GBDException
from gbd_core.util import eprint, confirm
from gbd_init.initializer import Initializer, InitializerException

try:
    from gbdc import extract_base_features, base_feature_names, extract_gate_features, gate_feature_names, isohash, wcnfisohash, wcnf_base_feature_names, extract_wcnf_base_features, opb_base_feature_names, extract_opb_base_features, checksani, checksani_feature_names
except ImportError:
    def extract_base_features(path, tlim, mlim):
        raise ModuleNotFoundError("gbdc not found", name="gbdc")
    
    def base_feature_names():
        return [ ]

    def extract_gate_features(path, tlim, mlim):
        raise ModuleNotFoundError("gbdc not found", name="gbdc")

    def gate_feature_names():
        return [ ]
    
    def isohash(path):
        raise ModuleNotFoundError("gbdc not found", name="gbdc")

    def extract_wcnf_base_features(path, tlim, mlim):
        raise ModuleNotFoundError("gbdc not found", name="gbdc")
    
    def wcnf_base_feature_names():
        return [ ]

    def extract_opb_base_features(path, tlim, mlim):
        raise ModuleNotFoundError("gbdc not found", name="gbdc")

    def opb_base_feature_names():
        return [ ]
    
    def checksani(path, tlim, mlim):
        raise ModuleNotFoundError("gbdc not found", name="gbdc")
    
    def checksani_feature_names():
        return [ ]

## GBDHash
def compute_hash(hash, path, limits):
    eprint('Hashing {}'.format(path))
    hash = identify(path)
    return [ ("local", hash, path), ("filename", hash, os.path.basename(path)) ]

## ISOHash
def compute_isohash(hash, path, limits):
    eprint('Computing ISOHash for {}'.format(path))
    context = get_context_by_suffix(path)
    if context == 'wcnf':
        ihash = wcnfisohash(path)
    else:
        ihash = isohash(path)
    return [ ('isohash', hash, ihash) ]

## Base Features
def compute_base_features(hash, path, limits, tp=None):
    eprint('Extracting base features from {} {}'.format(hash, path))
    rec = extract_base_features(path, limits['tlim'], limits['mlim'])
    return [ (key, hash, int(value) if isinstance(value, float) and value.is_integer() else value) for key, value in rec.items() ]

## Gate Features
def compute_gate_features(hash, path, limits, tp=None):
    eprint('Extracting gate features from {} {}'.format(hash, path))
    rec = extract_gate_features(path, limits['tlim'], limits['mlim'])
    return [ (key, hash, int(value) if isinstance(value, float) and value.is_integer() else value) for key, value in rec.items() ]   

## WCNF Base Features
def compute_wcnf_base_features(hash, path, limits, tp=None):
    eprint('Extracting WCNF base features from {} {}'.format(hash, path))
    rec = extract_wcnf_base_features(path, limits['tlim'], limits['mlim'])
    return [ (key, hash, int(value) if isinstance(value, float) and value.is_integer() else value) for key, value in rec.items() ]

## OPB Base Features
def compute_opb_base_features(hash, path, limits, tp=None):
    eprint('Extracting OPB base features from {} {}'.format(hash, path))
    rec = extract_opb_base_features(path, limits['tlim'], limits['mlim'])
    return [ (key, hash, int(value) if isinstance(value, float) and value.is_integer() else value) for key, value in rec.items() ]

## SANI Features
def compute_sani_features(hash, path, limits, tp=None):
    eprint('Extracting SANI features from {} {}'.format(hash, path))
    rec = checksani(path, limits['tlim'], limits['mlim'])
    return [ (key, hash, int(value) if isinstance(value, float) and value.is_integer() else value) for key, value in rec.items() ]


generic_extractors = {
    "base" : {
        "description" : "Extract base features from CNF files. ",
        "contexts" : [ "cnf", "sancnf" ],
        "features" : [ (name, "empty") for name in base_feature_names() ],
        "compute" : compute_base_features,
    },
    "checksani" : {
        "description" : "Extract sanitise status from CNF files. ",
        "contexts" : [ "cnf", "sancnf" ],
        "features" : [ (name, "empty") for name in checksani_feature_names() ],
        "compute" : compute_sani_features,
    },
    "gate" : {
        "description" : "Extract gate features from CNF files. ",
        "contexts" : [ "cnf", "sancnf" ],
        "features" : [ (name, "empty") for name in gate_feature_names() ],
        "compute" : compute_gate_features,
    },
    "isohash" : {
        "description" : "Compute ISOHash for CNF or WCNF files. ",
        "contexts" : [ "cnf", "wcnf", "sancnf" ],
        "features" : [ ("isohash", "empty") ],
        "compute" : compute_isohash,
    },
    "wcnfbase" : {
        "description" : "Extract base features from WCNF files. ",
        "contexts" : [ "wcnf" ],
        "features" : [ (name, "empty") for name in wcnf_base_feature_names() ],
        "compute" : compute_wcnf_base_features,
    },
    "opbbase" : {
        "description" : "Extract base features from OPB files. ",
        "contexts" : [ "opb" ],
        "features" : [ (name, "empty") for name in opb_base_feature_names() ],
        "compute" : compute_opb_base_features,
    }
}


def init_features_generic(key: str, api: GBD, rlimits, df, target_db):
    einfo = generic_extractors[key]
    context = api.database.dcontext(target_db)
    if not context in einfo["contexts"]:
        raise InitializerException("Target database context must be in {}".format(einfo["contexts"]))
    extractor = Initializer(api, rlimits, target_db, einfo["features"], einfo["compute"])
    extractor.create_features()
    extractor.run(df)


def init_local(api: GBD, rlimits, root, target_db):
    context = api.database.dcontext(target_db)
    
    features = [ ("local", None), ("filename", None) ]
    extractor = Initializer(api, rlimits, target_db, features, compute_hash)
    extractor.create_features()

    # Cleanup stale entries
    df = api.query(group_by=context + ":local")
    dfilter = df["local"].apply(lambda x: not x or not os.path.isfile(x))
    missing = df[dfilter]
    if len(missing) and api.verbose:
        for path in missing["local"].tolist():
            eprint(path)
    if len(missing) and confirm("{} files not found. Remove stale entries from local table?".format(len(missing))):
        api.reset_values("local", values=missing["local"].tolist())

    # Create df with paths not yet in local table
    paths = [ path for suffix in suffixes(context) for path in glob.iglob(root + "/**/*" + suffix, recursive=True) ]
    df2 = pd.DataFrame([(None, path) for path in paths if not path in df["local"].to_list()], columns=["hash", "local"])
    
    extractor.run(df2)


import toml
import os
import importlib.resources as pkg_resources

### Default Context
default = "cnf"

### Load Configuration from Files
def load_config(default_config_path, user_config_path=None):
    # Load the default configuration file
    with pkg_resources.open_text('gbd_tools', default_config_path) as f:
        config = toml.load(f)
    
    # If a user configuration file is provided, load it and update the config
    if user_config_path and os.path.exists(user_config_path):
        with open(user_config_path, 'r') as f:
            user_config = toml.load(f)
        config.update(user_config)
    
    return config

### Convert ConfigParser to Dictionary
def config_to_dict(config):
    config_dict = {}
    for context, details in config['contexts'].items():
        config_dict[context] = {
            "description": details["description"],
            "suffix": details["suffix"],
            "idfunc": globals()[details["idfunc"]],
        }
    return config_dict

### Paths to Configuration Files
default_config_path = "default_config.toml"
user_config_path = "user_config.toml"  # Adjust this path as needed

### Load and Convert Configuration
config_parser = load_config(default_config_path, user_config_path)
config = config_to_dict(config_parser)



import os
import unittest
import sqlite3

from gbd_core.api import GBD, GBDException
from gbd_core.schema import Schema

from tests import util

class APITestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.file1 = util.get_random_unique_filename('test1', '.db')
        self.file2 = util.get_random_unique_filename('test2', '.db')
        sqlite3.connect(self.file1).close()
        sqlite3.connect(self.file2).close()
        self.name1 = Schema.dbname_from_path(self.file1)
        self.name2 = Schema.dbname_from_path(self.file2)
        self.api = GBD([self.file1, self.file2])
        return super().setUp()

    def tearDown(self) -> None:
        if os.path.exists(self.file1):
            os.remove(self.file1)
        if os.path.exists(self.file2):
            os.remove(self.file2)
        return super().tearDown()

    def test_databases_exist(self):
        self.assertEquals(self.api.get_databases(), [ self.name1, self.name2 ])
        self.assertEquals(self.api.get_database_path(self.name1), self.file1)
        self.assertEquals(self.api.get_database_path(self.name2), self.file2)

    def test_create_feature(self):
        self.api.create_feature("A", None, self.name1)
        self.assertTrue(self.api.feature_exists("A"))
        self.api.create_feature("A", None, self.name2)
        api2 = GBD([self.file2])
        self.assertTrue(api2.feature_exists("A"))
        with self.assertRaises(GBDException):
            self.api.create_feature("A", None, self.name1)
        with self.assertRaises(GBDException):
            self.api.create_feature("A", None, self.name2)

    def test_delete_feature(self):
        self.api.create_feature("A", None, self.name1)
        self.api.create_feature("A", None, self.name2)
        self.api.delete_feature("A", self.name1)
        self.assertFalse(self.api.feature_exists("A", self.name1))
        self.assertTrue(self.api.feature_exists("A"))
        self.assertTrue(self.api.feature_exists("A", self.name2))
        self.api.delete_feature("A")
        self.assertFalse(self.api.feature_exists("A"))

    def test_rename_feature(self):
        self.api.create_feature("A", None, self.name1)
        self.api.create_feature("B", None, self.name1)
        self.api.create_feature("A", None, self.name2)
        self.api.rename_feature("A", "B", self.name2)
        self.assertFalse(self.api.feature_exists("A", self.name2))
        self.assertTrue(self.api.feature_exists("B", self.name2))
        self.assertTrue(self.api.feature_exists("A", self.name1))
        self.assertTrue(self.api.feature_exists("B", self.name1))
        with self.assertRaises(GBDException):
            self.api.rename_feature("A", "B", self.name1)

    def test_set_values(self):
        self.api.create_feature("A", None, self.name1) # feature is multi-valued
        self.api.create_feature("B", "empty", self.name1) # feature has default value
        self.api.create_feature("A", "empty", self.name2) # shadowed feature
        # value1 (set values, default values emerge)
        self.api.set_values("A", "value1", [ str(i) for i in range(100) ], self.name1)
        df = self.api.query("A = value1", resolve=["A", "B"])
        self.assertCountEqual(df['hash'].tolist(), [ str(i) for i in range(100) ])
        self.assertCountEqual(df['A'].tolist(), [ "value1" for _ in range(100) ])
        self.assertCountEqual(df['B'].tolist(), [ "empty" for _ in range(100) ])
        # value2 (set values, feature is multi-valued)
        self.api.set_values("A", "value2", [ str(i) for i in range(50) ], self.name1)
        df = self.api.query("A = value1 or A = value2", resolve=["A"], collapse=None)
        self.assertCountEqual(df['A'].tolist(), [ "value2" for _ in range(50) ] + [ "value1" for _ in range(100) ])
        # value3 (set values of shadowed feature by specifying target-database)
        self.api.set_values("A", "value3", [ str(i) for i in range(50) ], self.name2)
        df = self.api.query("A = value1 or A = value2", resolve=["A"], collapse=None)
        self.assertCountEqual(df['A'].tolist(), [ "value2" for _ in range(50) ] + [ "value1" for _ in range(100) ])
        self.api.database.commit()
        api2 = GBD([self.file2])
        df = api2.query("A = value3", resolve=["A"])
        self.assertCountEqual(df["A"].tolist(), [ "value3" for _ in range(50) ])

    def test_reset_values(self):
        self.api.create_feature("A", None, self.name1)
        self.api.create_feature("B", "empty", self.name1)
        self.api.create_feature("A", "empty", self.name2)
        self.api.set_values("A", "value1", [ str(i) for i in range(100) ], self.name1)
        self.api.set_values("A", "value2", [ str(i) for i in range(100) ], self.name1)
        self.api.set_values("B", "value3", [ str(i) for i in range(100) ], self.name1)
        self.api.set_values("A", "value1", [ str(i) for i in range(100) ], self.name2)
        # reset values in A
        self.api.reset_values("A", [ "value1" ], [ str(i) for i in range(50) ], self.name1)
        df = self.api.query(None, hashes=[ str(i) for i in range(100) ], resolve=["A"], collapse=None)
        self.assertCountEqual(df['A'].tolist(), [ "value1" for _ in range(50) ] + [ "value2" for _ in range(100) ])
        # reset values in B
        self.api.reset_values("B", [ "value3" ], [ str(i) for i in range(50) ], self.name1)
        df = self.api.query(None, hashes=[ str(i) for i in range(100) ], resolve=["B"])
        self.assertCountEqual(df['B'].tolist(), [ "value3" for _ in range(50) ] + [ "empty" for _ in range(50) ])
        # reset values in shadowed A
        self.api.database.verbose = True
        self.api.reset_values("A", [ "value1" ], [ str(i) for i in range(50) ], self.name2)
        self.api.database.commit()
        api2 = GBD([self.file2])
        df = api2.query("A = value1", resolve=["A"])
        self.assertCountEqual(df["A"].tolist(), [ "value1" for _ in range(50) ])


#!/usr/bin/python3

from gbd_core.api import GBD
from gbd_core import util
from gbd_core.util_argparse import *

from graph_tool.all import *
from pysat.formula import CNF

import itertools
import traceback
import os
import sys


def gen_vig(path):
    g = Graph(directed=False)
    cnf = CNF()
    cnf.from_file(path, compressed_with='lzma')
    print("NVars: {}".format(cnf.nv))
    g.add_vertex(cnf.nv + 1)
    w = g.new_edge_property("float")
    for clause in cnf:
        inc = 1.0 / pow(2, len(clause))
        for tup in itertools.combinations(clause, 2):
            e = g.edge(g.vertex(abs(tup[0])), g.vertex(abs(tup[1])), add_missing=True)
            w[e] = w[e] + inc
    return (g, w)


def vig(api: GBD, path, proof):
    (g, w) = gen_vig(path)
    pos = sfdp_layout(g, eweight=w)
    graph_draw(g, pos, output_size=(1000, 1000), vertex_color=[1,1,1,0],
           vertex_size=1, edge_pen_width=1.2, output="graph.pdf")

def read_by_tokens(fileobj):
    for line in fileobj:
        for token in line.split():
            yield token

from gi.repository import Gtk, Gdk, GdkPixbuf, GObject, GLib

def animate_proof(api: GBD, path, proof):
    (g, w) = gen_vig(path)
    pos = sfdp_layout(g, eweight=w)
    win = GraphWindow(g, pos, geometry=(800, 600))

    handle = open(proof)        
    def update(handle):
        delete = False
        clause = []
        sfdp_layout(g, pos=pos, K=0.5 , init_step=0.005, max_iter=1)
        for token in read_by_tokens(handle):
            if token == "d":
                delete = True
            elif token == "0":
                inc = 1.0 / pow(2, len(clause))
                if delete:
                    for tup in itertools.combinations(clause, 2):
                        e = g.edge(g.vertex(abs(tup[0])), g.vertex(abs(tup[1])))
                        w[e] = w[e] - inc
                else:
                    for tup in itertools.combinations(clause, 2):
                        e = g.edge(g.vertex(abs(tup[0])), g.vertex(abs(tup[1])))
                        w[e] = w[e] + inc
                print(clause)
                win.graph.regenerate_surface()
                win.graph.queue_draw()
                return True
            else:
                clause = clause + [int(token)]
        return False
        
    cid = GLib.idle_add(update, handle)
   
    # We will give the user the ability to stop the program by closing the window.

    win.connect("delete_event", Gtk.main_quit)


    # Actually show the window, and start the main loop.

    win.show_all()

    Gtk.main()


def cli_graph(api: GBD, args):
    from gbd_apps import graph
    graph.animate_proof(api, args.path, args.proof)



### Define Command-Line Interface and Map Sub-Commands to Methods
def main():
    parser = get_gbd_argparser()

    subparsers = parser.add_subparsers(help='Available Commands:', required=True, dest='gbd command')

    # GRAPHS
    parser_graph = subparsers.add_parser('graph', help='Visualize Formula')
    parser_graph.add_argument('path', type=file_type, help='CNF File')
    parser_graph.add_argument('proof', type=file_type, help='Proof File')
    parser_graph.set_defaults(func=cli_graph)

    # PARSE ARGUMENTS
    args = parser.parse_args()
    try:
        if hasattr(args, 'hashes') and not sys.stdin.isatty():
            if not args.hashes or len(args.hashes) == 0:
                args.hashes = util.read_hashes()  # read hashes from stdin
        with GBD(args.db.split(os.pathsep), args.verbose) as api:
            args.func(api, args)
    except Exception as e:
        util.eprint("{}: {}".format(type(e), str(e)))
        if args.verbose:
            util.eprint(traceback.format_exc())
        sys.exit(1)


if __name__ == '__main__':
    main()



#!/usr/bin/python3
from gbd_core.api import GBD
from gbd_core import util
from gbd_core.util_argparse import *

import os
import pandas as pd
from numpy.core.numeric import NaN

import matplotlib
matplotlib.use('GTK3Agg')
import matplotlib.pyplot as plt
import itertools

import traceback
import sys

#matplotlib.use("pgf")
#matplotlib.rcParams.update({
#    "pgf.texsystem": "pdflatex",
#    'font.family': 'serif',
#    'text.usetex': True,
#    'pgf.rcfonts': False,
#})

coolors=['#264653', '#2a9d8f', '#e9c46a', '#f4a261', '#e76f51']
coolors2=["#001219","#9b2226","#005f73","#ca6702","#0a9396","#bb3e03","#94d2bd","#ae2012","#e9d8a6","#ee9b00"]


def scatter(api: GBD, query, runtimes, timeout, groups):
    plt.rcParams.update({'font.size': 6})
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.set_aspect('equal', adjustable='box')
    plt.axline((0, 0), (1, 1), linewidth=0.5, color='grey', zorder=0)
    plt.axhline(y=timeout, xmin=0, xmax=1, linewidth=0.5, color='grey', zorder=0)
    plt.axvline(x=timeout, ymin=0, ymax=1, linewidth=0.5, color='grey', zorder=0)
    plt.xlabel(runtimes[0], fontsize=8)
    plt.ylabel(runtimes[1], fontsize=8)
    markers = itertools.cycle(plt.Line2D.markers.items())
    next(markers)
    next(markers)
    plt.rcParams['axes.prop_cycle'] = plt.cycler(color=coolors)
    if not groups:
        groups = []

    dfall = api.query(query, [], runtimes)
    for r in runtimes:
        dfall[r] = pd.to_numeric(dfall[r], errors='coerce')
        dfall.loc[(dfall[r] >= timeout) | pd.isna(dfall[r]), r] = timeout
    print(dfall)

    plots = []
    title = []
    for g in groups:
        color=next(ax._get_lines.prop_cycler)['color']
        marker=next(markers)[0]
        df = api.query(query + " and (" + g + ")", [], runtimes)
        for r in runtimes:
            df[r] = pd.to_numeric(df[r], errors='coerce')
            df.loc[(df[r] >= timeout) | pd.isna(df[r]), r] = timeout
        dfall = pd.concat([dfall, df]).drop_duplicates(keep=False)

        plots = plots + [ plt.scatter(data=df, x=runtimes[0], y=runtimes[1], c=color, marker=marker, alpha=0.7, linewidth=0.7, zorder=2) ]
        title = title + [ g ]

    plt.scatter(data=dfall, x=runtimes[0], y=runtimes[1], marker='.', alpha=0.7, linewidth=0.7, color="black", zorder=1)

    plt.legend(tuple(plots), tuple(title), scatterpoints=1, bbox_to_anchor=(0., 1.02, 1., .102), loc='lower left', ncol=5, mode="expand", borderaxespad=0.)
    plt.savefig('out.svg', transparent=True, bbox_inches='tight', pad_inches=0)
    plt.show()


def cdf(api: GBD, query, runtimes, timeout, title):
    plt.rcParams.update({'font.size': 8})
    df = api.query(query, [], runtimes)
    df[df == 'timeout'] = timeout
    df[df == 'empty'] = timeout
    df[runtimes] = df[runtimes].astype('float')
    df.drop('hash', axis=1, inplace=True)
    df[df >= timeout] = timeout
    df['vbs'] = df[runtimes].min(axis=1)
    print(df)

    plt.rcParams['axes.prop_cycle'] = plt.cycler(color=coolors2+coolors)

    params = {'legend.fontsize': 'small',
            'axes.labelsize': 6,
            'axes.titlesize': 6,
            'xtick.labelsize': 6,
            'ytick.labelsize': 6,
            'axes.titlepad': 10}
    plt.rcParams.update(params)

    fig = plt.figure()
    ax = fig.add_subplot(111)

    plt.xlim(0, timeout + 100)
    plt.grid(linestyle='-', linewidth=.5)
    plt.axvline(x=timeout, linestyle='dashed', color='black', linewidth=.5)
    #plt.ylim(0, len(result))

    # Build Title
    if (title is None):
        title = []
        for elem in query.split('=')[1:]:
            parts = elem.strip().split(' ')
            title = title + [parts[0].replace('_', ' ').title()]
        ax.set_title(", ".join(title), fontsize=6, variant='small-caps')
    else:
        ax.set_title(title, fontsize=6, variant='small-caps')

    df2 = pd.DataFrame(index=range(timeout+100), columns=runtimes)
    df2.fillna(0)
    for col in ['vbs'] + runtimes:
        df2[col] = [0] * (timeout + 100)
        for val in df[col]:
            if val < timeout:
                df2.loc[round(val), col] = df2[col][round(val)] + 1

        sum = 0
        for val in range(1, timeout + 100):
            df2.loc[val, col + "_"] = NaN
            if df2[col][val] != 0:
                df2.loc[val, col + "_"] = sum
            sum = sum + df2.loc[val, col]
            df2.loc[val, col] = sum
    
    markers = itertools.cycle(['o','v','^','<','>','p','P','*','h','H','8','X','d','D','s'])
    #next(markers)
    #next(markers)
    order=len(runtimes)+1
    for col in ['vbs'] + runtimes:
        color=next(ax._get_lines.prop_cycler)['color']
        ax.plot(df2[col], zorder=order, linestyle='-', linewidth=.5, color=color)
        ax.plot(df2[str(col) + "_"], label=col, zorder=order, fillstyle='none', marker=next(markers)[0], alpha=.9, markeredgewidth=.5, markersize=3, drawstyle='steps-post', color=color)
        order = order - 1
    
    plt.legend(ncol=2, loc='lower right')
    plt.savefig('out.svg', transparent=True, bbox_inches='tight', pad_inches=0)
    plt.show()

    # pgf output:
    #plt.savefig('out.pgf')

def cli_plot_scatter(api: GBD, args):
    from gbd_apps import plot
    plot.scatter(api, args.query, args.runtimes, args.tlim, args.groups)

def cli_plot_cdf(api: GBD, args):
    from gbd_apps import plot
    plot.cdf(api, args.query, args.runtimes, args.tlim, args.title)


### Define Command-Line Interface and Map Sub-Commands to Methods
def main():
    parser = get_gbd_argparser()

    subparsers = parser.add_subparsers(help='Available Commands:', required=True, dest='gbd command')

    parser_scatter = subparsers.add_parser('scatter', help='Scatter Plot')
    add_query_and_hashes_arguments(parser_scatter)
    parser_scatter.add_argument('-r', '--runtimes', help='Two runtime features', nargs=2)
    parser_scatter.add_argument('-g', '--groups', help='Highlight specific groups (e.g. family=cryptography)', nargs='+')
    parser_scatter.set_defaults(func=cli_plot_scatter)

    parser_cdf = subparsers.add_parser('cdf', help='CDF Plot')
    add_query_and_hashes_arguments(parser_cdf)
    parser_cdf.add_argument('-r', '--runtimes', help='List of runtime features', nargs='+')
    parser_cdf.add_argument('--title', help='Plot Title')
    parser_cdf.set_defaults(func=cli_plot_cdf)

    # PARSE ARGUMENTS
    args = parser.parse_args()
    try:
        if hasattr(args, 'hashes') and not sys.stdin.isatty():
            if not args.hashes or len(args.hashes) == 0:
                args.hashes = util.read_hashes()  # read hashes from stdin
        with GBD(args.db.split(os.pathsep), args.verbose) as api:
            args.func(api, args)
    except Exception as e:
        util.eprint("{}: {}".format(type(e), str(e)))
        if args.verbose:
            util.eprint(traceback.format_exc())
        sys.exit(1)


if __name__ == '__main__':
    main()



# MIT License

# Copyright (c) 2023 Markus Iser, Karlsruhe Institute of Technology (KIT)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

import sqlite3
import typing

from pprint import pprint

from gbd_core.util import eprint
from gbd_core.schema import Schema, FeatureInfo
from gbd_core import contexts


class DatabaseException(Exception):
    pass


class Database:

    def __init__(self, path_list: list, verbose=False, autocommit=True):
        self.verbose = verbose
        self.schemas = self.init_schemas(path_list)
        self.features = self.init_features()
        self.connection = sqlite3.connect("file::memory:?cache=shared", uri=True, timeout=10)
        self.cursor = self.connection.cursor()
        self.maindb = None
        self.autocommit = autocommit
        schema: Schema
        for schema in self.schemas.values():
            if not schema.is_in_memory():
                self.execute("ATTACH DATABASE '{}' AS {}".format(schema.path, schema.dbname))
            else:
                self.execute("ATTACH DATABASE 'file:{}?mode=memory&cache=shared' AS {}".format(schema.dbname, schema.dbname))
            # first database is the default database:
            if not self.maindb:
                self.maindb = schema.dbname
        
    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.connection.commit()
        self.connection.close()


    # returns major version of sqlite3 as float
    @classmethod
    def sqlite3_version(cls):
        return float(sqlite3.sqlite_version.rsplit('.', 1)[0])


    def init_schemas(self, path_list) -> typing.Dict[str, Schema]:
        result = dict()
        for path in path_list:
            schema = Schema.create(path)
            if not schema.dbname in result:
                result[schema.dbname] = schema
            elif schema.is_in_memory():
                result[schema.dbname].absorb(schema)
            else:
                raise DatabaseException("Database name collision on " + schema.dbname)
        return result


    # return a dictionary which maps feature names to feature infos
    def init_features(self) -> typing.Dict[str, FeatureInfo]:
        result = dict()
        schema: Schema
        for schema in self.schemas.values():
            feature: FeatureInfo
            for feature in schema.features.values():
                # first found feature is used: (=feature precedence by database position)
                if not feature.name in result:
                    result[feature.name] = [ feature ]
                elif feature.column == "hash" and feature.table == "features": 
                    # first found features table is the one that serves the hash
                    if result[feature.name][0].table != "features":
                        result[feature.name].insert(0, feature)
                    else:
                        result[feature.name].append(feature)
                else:
                    result[feature.name].append(feature)
        return result


    def query(self, q):
        if self.verbose:
            eprint(q)
        return self.cursor.execute(q).fetchall()

    def execute(self, q):
        if self.verbose:
            eprint(q)
        self.cursor.execute(q)
        if self.autocommit:
            self.commit()

    def commit(self):
        self.connection.commit()

    def set_auto_commit(self, autocommit):
        self.autocommit = autocommit


    def dexists(self, dbname):
        return dbname in self.schemas.keys()

    def dmain(self, dbname):
        return dbname == self.maindb

    def dpath(self, dbname):
        if not dbname in self.schemas:
            raise DatabaseException("Database '{}' not found".format(dbname))
        return self.schemas[dbname].path
        
    def dcontext(self, dbname):
        if not dbname in self.schemas:
            raise DatabaseException("Database '{}' not found".format(dbname))
        return self.schemas[dbname].context
        
    def dtables(self, dbname):
        if not dbname in self.schemas:
            raise DatabaseException("Database '{}' not found".format(dbname))
        return self.schemas[dbname].get_tables()


    def finfo(self, fname, db=None):
        if fname in self.features and len(self.features[fname]) > 0:
            if db is None:
                return self.features[fname][0]
            else:
                infos = [ info for info in self.features[fname] if info.database == db ]
                if len(infos) == 0:
                    raise DatabaseException("Feature '{}' does not exists in database {}".format(fname, db))
                return infos[0]
        else:
            raise DatabaseException("Feature '{}' does not exists".format(fname))
        

    def faddr_column(self, feature):
        finfo = self.find(feature)
        return "{}.{}.{}".format(finfo.database, finfo.table, finfo.column)

    def faddr_table(self, feature):
        finfo = self.find(feature)
        return "{}.{}".format(finfo.database, finfo.table)
    

    def find(self, fid: str, db: str=None):
        """ Find feature by name or feature identifier
        
            Args:
            fid: feature identifier, of the form "database:feature", "context:feature" or "feature"
            db: database name (optional), if given fid is unique without database: or context: prefix

            Returns:
            FeatureInfo object: the info object for the first found feature
            feature precedence is according to the order of databases in the path list
            ambiguity can be resolved by using one of the following methods.
            - by giving a database name as the second argument or
            - by using the fid syntax "database:feature"
            - by using the fid syntax "context:feature" (note that this does not necessarily resolve all ambiguity)

            Raises:
            DatabaseException: if feature is not found or given database info is ambiguous
        """
        parts = fid.split(":")
        if db is not None:
            if len(parts) > 1:
                if parts[0] != db:
                    raise DatabaseException("Ambiguous database identifiers: '{}' and '{}'".format(parts[0], db))
                else:
                    return self.finfo(parts[1], parts[0])
            return self.finfo(fid, db)
        elif len(parts) == 1:
            return self.finfo(fid)
        elif parts[0] in self.get_databases():
            return self.finfo(parts[1], parts[0])
        elif parts[0] in self.get_contexts():
            db = self.get_databases(parts[0])[0]
            return self.finfo(parts[1], db)
        else:
            raise DatabaseException("Feature '{}' not found".format(fid))
        

    def faddr(self, fid: str, with_column=True):
        finfo = self.find(fid)

        if with_column:
            return "{}.{}.{}".format(finfo.database, finfo.table, finfo.column)
        else:
            return "{}.{}".format(finfo.database, finfo.table)


    def get_databases(self, context: str=None):
        return [ dbname for (dbname, schema) in self.schemas.items() if not context or context == schema.context ]

    def get_contexts(self, dbs=[]):
        return list(set([ s.context for s in self.schemas.values() if not dbs or s.dbname in dbs ]))
  
    def get_features(self, dbs=[]):
        return [ name for (name, infos) in self.features.items() for info in infos if not dbs or info.database in dbs ]

    def get_tables(self, dbs=[]):
        tables = [ info.table for infos in self.features.values() for info in infos if not dbs or info.database in dbs ]
        return list(set(tables))


    def create_feature(self, name, default_value=None, target_db=None, permissive=False):
        db = target_db or self.maindb
        created = self.schemas[db].create_feature(name, default_value, permissive)
        for finfo in created:
            if not finfo.name in self.features.keys():
                self.features[finfo.name] = [ finfo ]
            else:
                # this code disregards feature precedence by database position:
                self.features[finfo.name].append(finfo)


    def set_values(self, fname, value, hashes, target_db=None):
        finfo = self.finfo(fname, target_db)
        self.schemas[finfo.database].set_values(fname, value, hashes)


    def rename_feature(self, fname, new_fname, target_db=None):
        Schema.valid_feature_or_raise(new_fname)
        finfo = self.finfo(fname, target_db)
        self.execute("ALTER TABLE {}.features RENAME COLUMN {} TO {}".format(finfo.database, fname, new_fname))
        if finfo.default is None:
            con = sqlite3.connect(self.schemas[finfo.database].path)
            with con as cursor:
                cursor.execute("ALTER TABLE {} RENAME TO {}".format(fname, new_fname))
            con.close()
        self.features[fname].remove(finfo)
        if not len(self.features[fname]):
            del self.features[fname]
        finfo.name = new_fname
        if not new_fname in self.features.keys():
            self.features[new_fname] = [ finfo ]
        else:
            # this code disregards feature precedence by database position:
            self.features[new_fname].append(finfo)


    def delete_feature(self, fname, target_db=None):
        finfo = self.finfo(fname, target_db)
        if finfo.default is None:
            self.execute('DROP TABLE IF EXISTS {}.{}'.format(finfo.database, fname))
        elif Database.sqlite3_version() >= 3.35:
            self.execute("ALTER TABLE {}.{} DROP COLUMN {}".format(finfo.database, finfo.table, fname))
        else:
            raise DatabaseException("Cannot delete unique feature {} with SQLite versions < 3.35".format(fname))
        self.features[fname].remove(finfo)
        if not len(self.features[fname]):
            del self.features[fname]


    def delete(self, fname, values=[], hashes=[], target_db=None):
        finfo = self.finfo(fname, target_db)
        w1 = "{cl} IN ('{v}')".format(cl=finfo.column, v="', '".join(values))
        w2 = "hash IN ('{h}')".format(h="', '".join(hashes))
        where = "{} AND {}".format(w1 if len(values) else "1=1", w2 if len(hashes) else "1=1")
        db = finfo.database
        if finfo.default is None:
            hashlist = [ r[0] for r in self.query("SELECT DISTINCT(hash) FROM {d}.{tab} WHERE {w}".format(d=db, tab=fname, w=where)) ]
            self.execute("DELETE FROM {d}.{tab} WHERE {w}".format(d=db, tab=fname, w=where))
            remaining = [ r[0] for r in self.query("SELECT DISTINCT(hash) FROM {d}.{tab} WHERE hash in ('{h}')".format(d=db, tab=fname, h="', '".join(hashlist))) ]
            setnone = [ h for h in hashlist if not h in remaining ]
            self.execute("UPDATE {d}.features SET {col} = 'None' WHERE hash IN ('{h}')".format(d=db, col=fname, h="', '".join(setnone)))
        else:
            self.execute("UPDATE {d}.features SET {col} = '{default}' WHERE {w}".format(d=db, col=fname, default=finfo.default, w=where))


    def delete_hashes_entirely(self, hashes, target_db=None):
        tables = self.get_tables([ target_db ])
        for table in tables:
            self.execute("DELETE FROM {}.{} WHERE hash IN ('{h}')".format(target_db, table, h="', '".join(hashes)))


    def copy_feature(self, old_name, new_name, target_db, hashlist=[]):
        old_finfo = self.find(old_name)
        data = self.query("SELECT hash, {col} FROM {d}.{tab} WHERE hash IN ('{h}')".format(d=old_finfo.database, col=old_finfo.column, tab=old_finfo.table, h="', '".join(hashlist)))
        for (hash, value) in data:
            self.set_values(new_name, value, [hash], target_db)





import unittest

from gbd_core.grammar import Parser, ParserException

class SchemaTestCase(unittest.TestCase):

    def setUp(self) -> None:
        return super().setUp()

    def tearDown(self) -> None:
        return super().tearDown()

    def test_query_nesting(self):
        parser = Parser("a = 1")
        self.assertEqual(parser.get_features(), set(["a"]))
        parser = Parser("a = 1 and b = 2")
        self.assertEqual(parser.get_features(), set(["a", "b"]))
        parser = Parser("a = 1 and (b = 2 or c = 3)")
        self.assertEqual(parser.get_features(), set(["a", "b", "c"]))
        parser = Parser("(b = 2 or c = 3) and a = 1")
        self.assertEqual(parser.get_features(), set(["a", "b", "c"]))

    def test_query_string_constraints(self):
        parser = Parser("a = val1")
        self.assertEqual(parser.get_features(), set(["a"]))
        parser = Parser("a = val1 and b != val2")
        self.assertEqual(parser.get_features(), set(["a", "b"]))
        parser = Parser("a like val1")
        self.assertEqual(parser.get_features(), set(["a"]))
        parser = Parser("a like val%")
        self.assertEqual(parser.get_features(), set(["a"]))
        parser = Parser("a like %val")
        self.assertEqual(parser.get_features(), set(["a"]))
        parser = Parser("a like %val%")
        self.assertEqual(parser.get_features(), set(["a"]))
        parser = Parser("a like val% and b unlike val%")
        self.assertEqual(parser.get_features(), set(["a", "b"]))        
        with self.assertRaises(ParserException):
            parser = Parser("a = %val%")

    def test_query_arithmetic_constraints(self):
        parser = Parser("a = (1 + 2)")
        self.assertEqual(parser.get_features(), set(["a"]))
        parser = Parser("a = (1 - 2)")
        self.assertEqual(parser.get_features(), set(["a"]))
        parser = Parser("a = ((1 + 2) / b)")
        self.assertEqual(parser.get_features(), set(["a", "b"]))
        parser = Parser("a = (b)")
        self.assertEqual(parser.get_features(), set(["a", "b"]))
        parser = Parser("a = b")
        self.assertEqual(parser.get_features(), set(["a"]))

    def test_explicit_context(self):
        parser = Parser("c:a = 1")
        self.assertEqual(parser.get_features(), set(["c:a"]))
        parser = Parser("c:a = 1 and d:b = 2")
        self.assertEqual(parser.get_features(), set(["c:a", "d:b"]))


#!/usr/bin/python3

# MIT License

# Copyright (c) 2023 Markus Iser, Karlsruhe Institute of Technology (KIT)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

from logging.handlers import TimedRotatingFileHandler
import os
import re

import flask
import logging
import waitress
from werkzeug.middleware.proxy_fix import ProxyFix

from gbd_core.database import DatabaseException
from gbd_core.api import GBD, GBDException
from gbd_core.grammar import ParserException
from gbd_core.util import is_number
from gbd_core import contexts

app = flask.Flask(__name__)

def request_query(request):
    query = ""
    if "query" in request.values:
        query = request.values.get('query')
    elif len(request.args) > 0:
        query = " and ".join(["{}={}".format(key, value) for (key, value) in request.args.items() if key != "context"])
    return query

def request_database(request):
    if "selected_db" in request.values and request.values.get('selected_db') in app.config['dbnames']:
        dbname = request.values.get('selected_db')
        context = request_context(request)
        if dbname in [ GBD.get_database_name(c) for c in app.config['contextdbs'][context] ]:
            return dbname
        else:
            return GBD.get_database_name(app.config['contextdbs'][context][0])
    else:
        return app.config['dbnames'][0]

def request_page(request):
    return int(request.values.get('page')) if "page" in request.values else 0

def request_action(request):
    return request.values.get('action') if "action" in request.values else "default"

def request_context(request):
    return request.values.get('context') if "context" in request.values else contexts.default_context()


def query_to_name(query):
    return re.sub(r'[^\w]', '_', query) if query else "allinstances"


def error_response(msg, addr, errno=404):
    app.logger.error("{}: {}".format(addr, msg))
    return flask.Response(msg, status=errno, mimetype="text/plain")

def file_response(text_blob, filename, mimetype, addr):
    app.logger.info("{}: Sending generated file {}".format(addr, filename))
    return flask.Response(text_blob, mimetype=mimetype, headers={"Content-Disposition": "attachment; filename=\"{}\"".format(filename), "filename": filename})

def path_response(path, filename, mimetype, addr):
    app.logger.info("{}: Sending file {}".format(addr, path))
    return flask.send_file(path, as_attachment=True, download_name=filename, mimetype=mimetype)

def json_response(json_blob, msg, addr):
    app.logger.info("{}: {}".format(addr, msg))
    return flask.Response(json_blob, status=200, mimetype="application/json")

def page_response(context, query, database, page=0):
    with GBD(app.config['contextdbs'][context]) as gbd:
        start = page * 1000
        end = start + 1000
        error = None
        try:
            df = gbd.query(query, resolve=['{}:{}'.format(database, f) for f in app.config["features"][database]], collapse="GROUP_CONCAT")
        except GBDException as err:
            error = "GBDException: {}".format(str(err))
        except DatabaseException as err:
            error = "DatabaseException: {}".format(str(err))
        except ParserException as err:
            error = "ParserException: {}".format(str(err))
        except Exception as err:
            error = "An Unhandled Exception Occurred"
        return flask.render_template('index.html', 
            context=context,
            error=error,
            contexts=app.config['contexts'],
            query=query,
            query_name=query_to_name(query), 
            result=df.iloc[start:end, :].values.tolist() if error is None else [], 
            total=len(df.index) if error is None else 0,
            page=page,
            pages=int(len(df.index) / 1000)+1 if error is None else 0,
            selected=database, 
            features=app.config["features"][database],
            databases=[ gbd.get_database_name(db) for db in app.config["contextdbs"][context] ],
            action=request_action(flask.request))


# Returns main index page
@app.route("/", methods=['POST', 'GET'])
def quick_search():
    context = request_context(flask.request)
    query = request_query(flask.request)
    database = request_database(flask.request)
    context_databases = [ GBD.get_database_name(db) for db in app.config["contextdbs"][context] ]
    if not database in context_databases:
        database = context_databases[0]
    page = request_page(flask.request)
    return page_response(context, query, database, page)


# Generates a list of URLs. Given query (text field of POST form) is executed and the hashes of the result are resolved
# against the filename feature. Every filename is associated with a URL to enable flexible downloading of these files
@app.route("/getinstances/", methods=['POST', 'GET'])
@app.route("/getinstances", methods=['POST', 'GET'])
def get_url_file():
    context = request_context(flask.request)
    with GBD(app.config['contextdbs'][context]) as gbd:
        query = request_query(flask.request)
        try:
            df = gbd.query(query)
        except (GBDException, DatabaseException, ParserException) as err:
            return error_response("{}, {}".format(type(err), str(err)), flask.request.remote_addr, errno=500)
        if context == "cnf":
            content = "\n".join([ flask.url_for("get_file", hashvalue=val, _external=True) for val in df['hash'].tolist() ])
        else:
            content = "\n".join([ flask.url_for("get_file", hashvalue=val, context=context, _external=True) for val in df['hash'].tolist() ])
        return file_response(content, query_to_name(query) + ".uri", "text/uri-list", flask.request.remote_addr)


# Send database file
@app.route('/getdatabase/')
@app.route('/getdatabase')
@app.route('/getdatabase/<database>/')
@app.route('/getdatabase/<database>')
def get_database_file(database=None):
    dbname=database if database and database in app.config['dbnames'] else app.config['dbnames'][0]
    dbpath=app.config['dbpaths'][dbname]
    return path_response(dbpath, os.path.basename(dbpath), 'application/x-sqlite3', flask.request.remote_addr)


# Find the file corresponding to the hashvalue and send it to the client
@app.route('/file/<hashvalue>/')
@app.route('/file/<hashvalue>')
def get_file(hashvalue):
    context = request_context(flask.request)
    print(context, app.config['contextdbs'][context])
    with GBD(app.config['contextdbs'][context]) as gbd:
        df = gbd.query(hashes=[hashvalue], resolve=['local', 'filename'], collapse="MIN")
        if not len(df.index):
            return error_response("Hash '{}' not found".format(hashvalue), flask.request.remote_addr)
        row = df.iloc[0]
        if not os.path.exists(row['local']):
            return error_response("Files temporarily not accessible", flask.request.remote_addr)
        return path_response(row['local'], row['hash'] + "-" + row['filename'], 'application/x-xz', flask.request.remote_addr)


# start the server
def serve(gbd: GBD, port: int = 5000, logdir: str = "/tmp"):
    formatter = logging.Formatter(fmt='[%(asctime)s, %(name)s, %(levelname)s] %(module)s.%(filename)s.%(funcName)s():%(lineno)d\n%(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    logging.getLogger().setLevel(logging.DEBUG)
    # Add sys.stdout to logging output
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(formatter)
    console_handler.setLevel(logging.INFO)
    logging.getLogger().addHandler(console_handler)
    # Add handler to write in rotating logging files
    file_handler = TimedRotatingFileHandler(logdir + "/trfile.log", when="midnight", backupCount=10)
    file_handler.setFormatter(formatter)
    file_handler.setLevel(logging.WARNING)
    logging.getLogger().addHandler(file_handler)

    global app
    app.wsgi_app = ProxyFix(app.wsgi_app, x_for=1)
    
    app.jinja_env.trim_blocks = True
    app.jinja_env.lstrip_blocks = True
    
    app.jinja_env.tests['link_field'] = lambda field: field is not None and field.startswith("http")
    app.jinja_env.tests['num_field'] = lambda field: field is not None and is_number(field)
    app.jinja_env.tests['int_field'] = lambda field: field is not None and field.isnumeric()

    path = os.path.dirname(__file__)
    app.static_folder = os.path.join(path, "static")
    app.template_folder = os.path.join(path, "templates")

    app.config['contexts'] = gbd.get_contexts()
    app.config['dbnames'] = gbd.get_databases()
    # group databases by context
    app.config['contextdbs'] = dict()
    for ctxt in app.config['contexts']:
        app.config['contextdbs'][ctxt] = [ gbd.get_database_path(c) for c in gbd.get_databases(ctxt) ]
    # group features by database
    app.config['dbpaths'] = dict()
    app.config['features'] = dict()
    for db in app.config['dbnames']:
        app.config['features'][db] = [ f for f in gbd.get_features(db) if not f in [ "hash", "local" ] ]
        app.config['dbpaths'][db] = gbd.get_database_path(db)
    app.config['features_flat'] = [ f for f in gbd.get_features() if not f in [ "hash", "local" ] ]

    waitress.serve(app, host='0.0.0.0', port=port)



#!/usr/bin/python3

# MIT License

# Copyright (c) 2023 Markus Iser, Karlsruhe Institute of Technology (KIT)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

from gbd_core.api import GBD
from gbd_core import util
from gbd_core.util_argparse import *

from itertools import combinations
from operator import itemgetter
import traceback
import sys

import mip
import pandas as pd


def cli_eval_combinations_ilp(api: GBD, args):
    df = api.query(args.query, [], args.runtimes)
    df = df[args.runtimes][df.applymap(util.is_number)].applymap(float)
    df.fillna(2*args.tlim, inplace=True)
    model = mip.Model()
    instance_solver_vars = [[model.add_var(f'x_{i}_{j}', var_type=mip.BINARY)
                            for j in range(df.shape[1])] for i in range(df.shape[0])]
    solver_vars = [model.add_var(f's_{j}', var_type=mip.BINARY)for j in range(df.shape[1])]
    for var_list in instance_solver_vars:  # per-instance constraints
        model.add_constr(mip.xsum(var_list) == 1)
    for j in range(df.shape[1]):  # per-solver-constraints
        model.add_constr(mip.xsum(instance_solver_vars[i][j] for i in range(df.shape[0])) <= df.shape[0] * solver_vars[j])  # "Implies" in Z3
    model.add_constr(mip.xsum(solver_vars) <= args.size)
    model.objective = mip.minimize(mip.xsum(instance_solver_vars[i][j] * int(df.iloc[i, j])
                                            for i in range(df.shape[0]) for j in range(df.shape[1])))
    print("SBS Score: {}".format(df.sum().min() / len(df.index)))
    model.verbose = int(api.verbose)
    model.threads = api.jobs
    model.optimize()
    print("VBS Score (k={}): {}".format(args.size, model.objective_value / len(df.index)))
    for index, item in enumerate([var.x for var in solver_vars]):
        if item > 0:
            print(args.runtimes[index])

def cli_eval_combinations_greedy(api: GBD, args):
    df = api.query(args.query, [], args.runtimes)
    df[args.runtimes] = df[args.runtimes][df.applymap(util.is_number)].applymap(float)
    df.fillna(2*args.tlim, inplace=True)
    runtimes = ["dummy"] + args.runtimes
    for comb in combinations(range(1, len(runtimes)), args.size):
        comb_par2 = sum([min(itemgetter(*comb)(row)) for _, row in df.iterrows()]) / len(df.index)
        print(str(itemgetter(*comb)(runtimes)) + ": " + str(comb_par2))


### Define Command-Line Interface and Map Sub-Commands to Methods
def main():
    parser = get_gbd_argparser()

    subparsers = parser.add_subparsers(help='Available Commands:', required=True, dest='gbd command')

    parser_ilp = subparsers.add_parser('ilp', help='ILP Solution')
    add_query_and_hashes_arguments(parser_ilp)
    parser_ilp.add_argument('-r', '--runtimes', help='List of runtime features', nargs='+')
    parser_ilp.add_argument('-k', '--size', help='Portfolio Size', type=int, default=2)
    parser_ilp.set_defaults(func=cli_eval_combinations_ilp)

    parser_greedy = subparsers.add_parser('greedy', help='Greedy Solution')
    add_query_and_hashes_arguments(parser_greedy)
    parser_greedy.add_argument('-r', '--runtimes', help='List of runtime features', nargs='+')
    parser_greedy.add_argument('-k', '--size', help='Portfolio Size', type=int, default=2)
    parser_greedy.set_defaults(func=cli_eval_combinations_greedy)

    # PARSE ARGUMENTS
    args = parser.parse_args()
    try:
        if hasattr(args, 'hashes') and not sys.stdin.isatty():
            if not args.hashes or len(args.hashes) == 0:
                args.hashes = util.read_hashes()  # read hashes from stdin
        with GBD(args.db.split(os.pathsep), args.verbose) as api:
            args.func(api, args)
    except Exception as e:
        util.eprint("{}: {}".format(type(e), str(e)))
        if args.verbose:
            util.eprint(traceback.format_exc())
        sys.exit(1)


if __name__ == '__main__':
    main()


#!/usr/bin/python3

import numpy as np
import pandas as pd
import gbd_core.util as util
import piskle
import argparse
import os
import sys
import traceback

from sklearn import tree
from sklearn.model_selection import train_test_split
from sklearn.metrics import classification_report

from gbd_core.api import GBD
from gbd_core import util
from gbd_core.util_argparse import *

class GBDException(Exception):
    pass


def classify(api: GBD, query, feature, hashes, features, collapse, group_by, filename, flag):
    df = api.query(query + " and {f} != unknown and {f} != empty".format(f=feature), hashes, features + [ feature ], collapse, group_by)

    for (key, value) in [("timeout", np.nan), ("memout", np.nan)]:
        df.replace(key, value, inplace=True)

    df.drop(["hash"], axis=1, inplace=True)

    # right-hand side (of classifier)
    categories = df.pop(feature).astype("category")
    y = categories.cat.codes.to_numpy()

    # left-hand side (of classifier)
    x = np.nan_to_num(df.to_numpy().astype(np.float32))

    # create classificator based on data given
    if flag == 0:
        clf = tree.DecisionTreeClassifier()
        clf = clf.fit(x, y)
        piskle.dump(clf, filename + '.pskl')

    #apply existing classificator
    elif flag == 1:
        clf = piskle.load(filename)
        cl = clf.predict(x)
        class_df = pd.DataFrame(cl, columns = ['predicted'])
        print(classification_report(y, class_df))

    # 5 fold cross validation
    elif flag ==3:
        x_train, x_test, y_train, y_test = train_test_split(x, y, test_size=0.2, random_state=0)
        clf = tree.DecisionTreeClassifier()
        clf = clf.fit(x_train, y_train)
        cl = clf.predict(x_test)
        class_df = pd.DataFrame(cl, columns = ['predicted'])
        pd.set_option('display.max_rows', None)
        file = open(filename, 'w')
        file.writelines("Classification report: " + classification_report(y_test, class_df) + ".\n")
        file.close()


def cli_classify(api: GBD, args):
    from gbd_apps import classification
    classification.classify(api, args.query, args.feature, args.hashes, args.resolve, args.collapse, args.group_by, args.timeout_memout, args.save, args.mode)


def main():
    parser = get_gbd_argparser()
    
    add_query_and_hashes_arguments(parser)
    parser.add_argument('-f', '--feature', help='Feature that should be classified', required=True)
    parser.add_argument('-r', '--resolve', help='List of features to resolve against', nargs='+', required=True)
    parser.add_argument('-c', '--collapse', default='group_concat',
                                   choices=['group_concat', 'min', 'max', 'avg', 'count', 'sum', 'none'],
                                   help='Treatment of multiple values per hash (or grouping value resp.)')
    parser.add_argument('-s', '--save', help='Filename')
    parser.add_argument('-g', '--group_by', default='hash', help='Group by specified attribute value')
    parser.add_argument('-d', '--dict', default=[], help='Dictionary to replace the margin values')
    parser.add_argument('-m', '--mode', default ='0', help='How to evaluate the classification. 0: generates and stores classifier. 1: applies given classifier. 2: generates a classifier and evaluates it.' )
    parser.set_defaults(func=cli_classify)

    # PARSE ARGUMENTS
    args = parser.parse_args()
    try:
        if hasattr(args, 'hashes') and not sys.stdin.isatty():
            if not args.hashes or len(args.hashes) == 0:
                args.hashes = util.read_hashes()  # read hashes from stdin
        with GBD(args.db.split(os.pathsep), args.verbose) as api:
            args.func(api, args)
    except Exception as e:
        util.eprint("{}: {}".format(type(e), str(e)))
        if args.verbose:
            util.eprint(traceback.format_exc())
        sys.exit(1)


if __name__ == '__main__':
    main()




import unittest
import sqlite3
import os

from gbd_core.schema import Schema
from gbd_core.database import Database
from gbd_core.query import GBDQuery

import tests.util as util

class QueryNonUniqueTestCase(unittest.TestCase):

    feat = "nonuniquefeature"
    feat2 = "nonuniquefeature2"
    feat3 = "numericfeature"
    val1 = "value1"
    val2 = "value2"
    hashes = [ "a", "b", "c" ]

    def setUp(self) -> None:
        self.file1 = util.get_random_unique_filename('test1', '.db')
        self.file2 = util.get_random_unique_filename('test2', '.db')
        sqlite3.connect(self.file1).close()
        sqlite3.connect(self.file2).close()
        self.dbname1 = Schema.dbname_from_path(self.file1)
        self.dbname2 = Schema.dbname_from_path(self.file2)
        self.db = Database([self.file1,self.file2], verbose=False)

        self.db.create_feature(self.feat, default_value=None, target_db=self.dbname1)
        self.db.set_values(self.feat, self.val1, self.hashes)

        self.db.create_feature(self.feat, default_value=None, target_db=self.dbname2)
        self.db.set_values(self.feat, self.val1, self.hashes[:1], target_db=self.dbname2)
        self.db.set_values(self.feat, self.val2, self.hashes, target_db=self.dbname2)

        self.db.create_feature(self.feat2, default_value=None, target_db=self.dbname2)
        self.db.set_values(self.feat2, self.val2, self.hashes)

        self.db.create_feature(self.feat3, default_value=0, target_db=self.dbname1)
        self.db.set_values(self.feat3, 1, self.hashes[0])
        self.db.set_values(self.feat3, 10, self.hashes[1])
        self.db.set_values(self.feat3, 100, self.hashes[2])

        return super().setUp()

    def tearDown(self) -> None:
        if os.path.exists(self.file1):
            os.remove(self.file1)
        if os.path.exists(self.file2):
            os.remove(self.file2)
        return super().tearDown()
    
    def simple_query(self, feat, val, dbname=None):
        if dbname is None:
            return self.query("{}={}".format(feat, val))
        else:
            return self.query("{}:{}={}".format(dbname, feat, val))
    
    def query(self, query):
        q = GBDQuery(self.db, query).build_query()
        return [ hash for (hash, ) in self.db.query(q) ]

    def dump(self):
        import sqlite3
        conn = sqlite3.connect(self.file)
        for line in conn.iterdump():
            print(line)
        conn.close()


    def test_feature_precedence_rules(self):
        res = self.simple_query(self.feat, self.val1)
        self.assertEqual(len(res), 3)
        res = self.simple_query(self.feat, self.val2)
        self.assertEqual(len(res), 0)
        res = self.simple_query(self.feat, self.val1, self.dbname1)
        self.assertEqual(len(res), 3)
        res = self.simple_query(self.feat, self.val2, self.dbname2)
        self.assertEqual(len(res), 3)

    def test_string_inequality(self):
        res = self.query("{} < {}".format(self.feat, self.val2))
        self.assertEqual(len(res), 3)
        res = self.query("{} > {}".format(self.feat, self.val1))
        self.assertEqual(len(res), 0)

    def test_numeric_inequality(self):
        res = self.query("{} < 2".format(self.feat3))
        self.assertEqual(len(res), 1)
        
    def test_multivalued_subselect(self):
        res = self.query("{db}:{f} != {v1} and {db}:{f} = {v2}".format(f=self.feat, v1=self.val1, v2=self.val2, db=self.dbname2))
        self.assertEqual(len(res), 2)

    def test_feature_accessible(self):
        res = self.simple_query(self.feat2, self.val2)
        self.assertEqual(len(res), 3)


import random
import os

def get_random_clause(max_len=10, max_vars=30):
    return ' '.join([str(random.randint(-max_vars, max_vars)) for _ in range(random.randint(0, max_len))]) + ' 0'

def get_random_formula(max_num=50):
    return '\n'.join([get_random_clause() for _ in range(random.randint(0, max_num))]) + '\n'

def get_random_unique_filename(prefix='random', suffix='.cnf'):
    filename = prefix + suffix
    while os.path.exists(filename):
        filename = '{}{}{}'.format(prefix, random.randint(0, 1000), suffix)
    return filename

def get_random_cnffile(max_num=50):
    filename = get_random_unique_filename()
    with open(filename, 'w') as f:
        f.write('p cnf {} {}\n'.format(random.randint(1, 100), random.randint(1, 100)))
        f.write(get_random_formula(max_num))
    return filename


#!/usr/bin/python3

# MIT License

# Copyright (c) 2023 Markus Iser, Karlsruhe Institute of Technology (KIT)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

from gbd_core.api import GBD
import gbd_core.util as util
from gbd_core.util_argparse import *

import traceback
import sys

from functools import reduce

def par2(series, timeout, divisor):
    return reduce(lambda x, y: x + (y if y < timeout else 2*timeout) / divisor, series, 0)

def cli_eval_par2(api: GBD, args):
    df = api.query(args.query, [], args.runtimes)
    df[args.runtimes] = df[args.runtimes][df.applymap(util.is_number)].applymap(float)
    df['vbs'] = df[args.runtimes].min(axis=1)
    div = len(df.index) if args.divisor is None else args.divisor
    timeout = args.tlim
    scores = df[args.runtimes].agg(par2, axis=0, timeout=timeout, divisor=div)
    print(scores)

def cli_eval_vbs(api: GBD, args):
    df = api.query(args.query, [], args.runtimes)
    df['vbs'] = df[args.runtimes].min(axis=1)
    for index, row in df.iterrows():
        print(row['hash'], row['vbs'])


### Define Command-Line Interface and Map Sub-Commands to Methods
def main():
    parser = get_gbd_argparser()

    subparsers = parser.add_subparsers(help='Available Commands:', required=True, dest='gbd command')

    parser_par2 = subparsers.add_parser('par2', help='Calculate Penalized Average Runtimes')
    add_query_and_hashes_arguments(parser_par2)
    parser_par2.add_argument('-r', '--runtimes', help='List of runtime features', nargs='+')
    parser_par2.add_argument('-d', '--divisor', help='Divisor for PAR2', type=int, default=None)
    parser_par2.set_defaults(func=cli_eval_par2)

    parser_vbs = subparsers.add_parser('vbs', help='Calculate Virtual Best Solver')
    add_query_and_hashes_arguments(parser_vbs)
    parser_vbs.add_argument('-r', '--runtimes', help='List of runtime features', nargs='+')
    parser_vbs.set_defaults(func=cli_eval_vbs)

    # PARSE ARGUMENTS
    args = parser.parse_args()
    try:
        if hasattr(args, 'hashes') and not sys.stdin.isatty():
            if not args.hashes or len(args.hashes) == 0:
                args.hashes = util.read_hashes()  # read hashes from stdin
        with GBD(args.db.split(os.pathsep), args.verbose) as api:
            args.func(api, args)
    except Exception as e:
        util.eprint("{}: {}".format(type(e), str(e)))
        if args.verbose:
            util.eprint(traceback.format_exc())
        sys.exit(1)


if __name__ == '__main__':
    main()



# MIT License

# Copyright (c) 2023 Markus Iser, Karlsruhe Institute of Technology (KIT)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.


import os
from functools import reduce

from gbd_core import contexts
from gbd_core.api import GBD, GBDException
from gbd_core import util

from gbd_core.contexts import identify
from gbd_init.initializer import Initializer, InitializerException

try:
    from gbdc import cnf2kis, sanitise, normalise
except ImportError:
    def cnf2kis(ipath, opath):
        raise ModuleNotFoundError("gbdc not found", name="gbdc")
    
    def sanitise(ipath, opath):
        raise ModuleNotFoundError("gbdc not found", name="gbdc")
    
    def normalise(ipath, opath):
        raise ModuleNotFoundError("gbdc not found", name="gbdc")


def kis_filename(path):
    kispath = reduce(lambda path, suffix: path[:-len(suffix)] if path.endswith(suffix) else path, contexts.suffixes('cnf'), path)
    return kispath + '.kis'

def sanitised_filename(path):
    sanpath = reduce(lambda path, suffix: path[:-len(suffix)] if path.endswith(suffix) else path, contexts.suffixes('cnf'), path)
    return sanpath + '.sanitized.cnf'

def normalised_filename(path):
    normpath = reduce(lambda path, suffix: path[:-len(suffix)] if path.endswith(suffix) else path, contexts.suffixes('cnf'), path)
    return normpath + '.normalised.cnf'


def wrap_cnf2kis(hash, path, limits):
    kispath = kis_filename(path)
    util.eprint('Transforming {} to k-ISP {}'.format(path, kispath))
    try:
        result = cnf2kis(path, kispath)
        if "local" in result:
            kishash = result['hash']
            return [ ('local', kishash, result['local']), ('to_cnf', kishash, hash),
                        ('nodes', kishash, result['nodes']), ('edges', kishash, result['edges']), ('k', kishash, result['k']) ]
        else:
            raise GBDException("CNF2KIS failed for {} due to {}".format(path, result['hash']))
    except Exception as e:
        util.eprint(str(e))
        if os.path.exists(kispath):
            os.remove(kispath)

    return [ ]

def wrap_sanitise(hash, path, limits):
    sanpath = sanitised_filename(path)
    util.eprint('Sanitising {}'.format(path))
    try:
        with open(sanpath, 'w') as f, util.stdout_redirected(f):
            result = sanitise(path, sanpath)
            if "local" in result:
                sanhash = result['hash']
                return [ ('local', sanhash, result['local']), ('to_cnf', sanhash, hash) ]
            else:
                raise GBDException("Sanitization failed for {}".format(path))
    except Exception as e:
        util.eprint(str(e))
        if os.path.exists(sanpath):
            os.remove(sanpath)

    return [ ]

def wrap_normalise(hash, path, limits):
    normpath = normalised_filename(path)
    util.eprint('Normalising {}'.format(path))
    try:
        with open(normpath, 'w') as f, util.stdout_redirected(f):
            result = normalise(path, normpath)
            normhash = result['hash']
            if "local" in result and hash == normhash:
                return [ ('local', normhash, result['local']) ]
            else:
                raise GBDException("Normalisation failed for {}".format(path))
    except Exception as e:
        util.eprint(str(e))
        if os.path.exists(normpath):
            os.remove(normpath)

    return [ ]


def transform_instances_generic(key: str, api: GBD, rlimits, query, hashes, target_db, source):
    einfo = generic_transformers[key]
    context = api.database.dcontext(target_db)
    if not context in einfo["target"]:
        raise InitializerException("Target database context must be in {}".format(einfo["target"]))
    if not source in einfo["source"]:
        raise InitializerException("Source database context must be in {}".format(einfo["source"]))
    transformer = Initializer(api, rlimits, target_db, einfo["features"], einfo["compute"])
    transformer.create_features()

    df = api.query(query, hashes, [source+":local"], collapse=None)
    dfilter = df['local'].apply(lambda x: x and not os.path.isfile(einfo["filename"](x)))

    transformer.run(df[dfilter])


generic_transformers = {
    "sanitise" : {
        "description" : "Sanitise CNF files. ",
        "source" : [ "cnf" ],
        "target" : [ "sancnf" ],
        "features" : [ ('local', None) , ('to_cnf', None) ],
        "compute" : wrap_sanitise,
        "filename" : sanitised_filename,
    },
    "normalise" : {
        "description" : "Normalise CNF files. ",
        "source" : [ "cnf" ],
        "target" : [ "cnf" ],
        "features" : [ ('local', None) ],
        "compute" : wrap_normalise,
        "filename" : normalised_filename,
    },
    "cnf2kis" : {
        "description" : "Transform CNF files to k-ISP instances. ",
        "source" : [ "cnf" ],
        "target" : [ "kis" ],
        "features" : [ ('local', None), ('to_cnf', None), ('nodes', 'empty'), ('edges', 'empty'), ('k', 'empty') ],
        "compute" : wrap_cnf2kis,
        "filename" : kis_filename,
    },
}


import unittest
import random
import os

from gbd_core.contexts import identify
from tests import util


class TestGBDHash(unittest.TestCase):

    def setUp(self):
        self.reference = util.get_random_formula()
        self.ref_file = "reference.cnf"
        with open(self.ref_file, 'w') as ref:
            ref.write(self.reference)
        self.reference_hash = identify(self.ref_file)

    def tearDown(self):
        if self.currentResult.wasSuccessful():
            os.remove(self.ref_file)

    def run(self, result=None):
        self.currentResult = result
        unittest.TestCase.run(self, result)

    def get_random_character(self):
        c = chr(random.randint(0, 255))
        return c if not c.isspace() else ' '

    def get_random_string(self, min_length=0, max_length=20):
        return ''.join([self.get_random_character() for _ in range(random.randint(min_length, max_length))])

    def get_random_whitespace_character(self):
        r = random.random()
        return '\t' if r < 0.25 else '\r' if r < 0.5 else '\n' if r < 0.75 else ' '

    def get_random_whitespace(self, min_length=0, max_length=3):
        return ''.join([self.get_random_whitespace_character() for _ in range(random.randint(min_length, max_length))])
    
    def get_random_header(self, p=0.5):
        return "p cnf {} {}\n".format(random.randint(1, 100), random.randint(1, 100)) if random.random() < p else ""

    def get_random_comment(self, p=0.5):
        return "c {}\n".format(self.get_random_string()) if random.random() < p else ""

    def test_randomized_variants(self):
        for _ in range(100):
            variant = self.get_random_whitespace()
            variant += self.get_random_comment()
            variant += self.get_random_header()
            variant += self.get_random_whitespace()
            for c in self.reference:
                if c.isspace():
                    variant += self.get_random_whitespace()
                variant += c
                if c.isspace():
                    variant += self.get_random_whitespace()
            variant += self.get_random_whitespace()

            var_file = "variant.cnf"
            with open(var_file, 'w') as f:
                f.write(variant)
            variant_hash = identify(var_file)
            if self.reference_hash == variant_hash:
                os.remove(var_file)

            self.assertEqual(self.reference_hash, variant_hash)
